@@ -8,6 +8,10 @@ import psutil
 import tempfile
 from collections import defaultdict
 import os
+from functools import reduce, partial
+import pandas as pd
+import gzip
+from mimetypes import guess_type
 
 log = logging.getLogger("my_logger")
 
@@ -15,9 +19,50 @@ sys.setrecursionlimit(10 ** 6)
 
 # from https://softwareengineering.stackexchange.com/a/419985
 def line_estimation(filename, first_size=1 << 24):
-    with open(filename, "rb") as file:
+    encoding = guess_type(filename)[1]
+    _open = partial(gzip.open, mode="rb") if encoding == "gzip" else open
+    with _open(filename) as file:
         buf = file.read(first_size)
-        return len(buf) // buf.count(b"\n")
+        if buf.count(b"\n") == 0:
+            log.info("No lines found in file to process. Exiting...")
+            exit(0)
+        else:
+            return len(buf) // buf.count(b"\n")
+
+
+# def get_multi_keys(aln_rows, key_col):
+#     keys = aln_rows[0]
+#     aln_rows.pop(0)
+#     if len(aln_rows) > 0:
+#         for k in aln_rows:
+#             keys.rbind(k)
+#     keys = keys[
+#         :,
+#         {"n": dt.count()},
+#         dt.by(
+#             key_col,
+#         ),
+#     ]
+#     # keys_multi = keys[dt.f.n > 1, key_col]
+#     # keys_nomulti = keys[dt.f.n == 1, key_col]
+#     # keys_nomulti.key = key_col
+#     keys[:, dt.update(is_multi=dt.ifelse(dt.f.n > 1, "multi", "nomulti"))]
+#     keys.key = key_col
+#     return keys
+
+
+def filter_eval_perc(alns, evalues, perc=0.1):
+    logging.info(f"Removing alignments within {int(perc * 100)}% of the best e-value")
+    evalues[:, "maxlogEvalue"] = -1 * perc * (np.log(evalues[:, "eVal"]) / np.log(10))
+    evalues.key = "subjectId"
+    aln_sizes = []
+    for i in np.arange(len(alns)):
+        alns[i] = alns[i][:, :, dt.join(evalues)]
+        alns[i][:, "logEvalue"] = -1 * ((np.log(alns[i][:, "eVal"])) / np.log(10))
+        aln_sizes.append(alns[i][(dt.f.logEvalue > dt.f.maxlogEvalue), :].shape[0])
+        del alns[i]["maxlogEvalue"]
+        del alns[i]["logEvalue"]
+    return aln_sizes
 
 
 def read_and_filter_alns(
@@ -41,6 +86,8 @@ def read_and_filter_alns(
         "slen",
     ],
     threads=1,
+    evalue_perc=None,
+    evalue_perc_step=0.1,
 ):
     """Read and filter alignments.
 
@@ -61,14 +108,16 @@ def read_and_filter_alns(
     # in the order of billions.
 
     logging.info("Getting a raw estimate of the number of alignments...")
-    nalns = os.path.getsize(aln) // line_estimation(aln)
+    nalns = os.path.getsize(aln) // line_estimation(
+        filename=aln, first_size=int(os.path.getsize(aln) * 0.01)
+    )
     logging.info(f"Approximately {nalns:,} alignments found.")
 
     max_rows = int((2 ** 31) - 1)
-
+    # max_rows = int(5e6)
     # if the number of alignmentsis larger than the number of rows
     # that can be read by fread, we need to read it by chunks instead
-    alns = []
+
     if nalns > max_rows:
         gc.collect()
         logging.info(
@@ -79,7 +128,8 @@ def read_and_filter_alns(
         )
 
         k, m = divmod(nalns, max_rows)
-
+        evalues = []
+        alns = []
         for i in range(k + 1):
             logging.info(f"::: Reading chunk #{i+1:,}.")
             up = max_rows * (i + 1)
@@ -107,14 +157,82 @@ def read_and_filter_alns(
             logging.info(
                 f"::: Filtering alignments in chunk #{i+1:,} with bitscore >= {bitscore} and evalue <= {evalue}"
             )
-            alns.append(
-                aln_chunk[(dt.f.eVal < evalue) & (dt.f.bitScore > bitscore), :].copy(
-                    deep=True
-                )
-            )
+            aln_chunk = aln_chunk[(dt.f.eVal < evalue) & (dt.f.bitScore > bitscore), :]
+            logging.info(f"Getting evalues")
+            evals = aln_chunk[:1, ["eVal"], dt.by(dt.f.subjectId), dt.sort(dt.f.eVal)]
+            # evals[:, "maxlogEvalue"] = (
+            #     -1 * evalue_perc * (np.log(evals[:, "eVal"]) / np.log(10))
+            # )
+            logging.info(f"Deep copying resulting frames")
+            evalues.append(evals.copy(deep=True))
+            alns.append(aln_chunk.copy(deep=True))
             del aln_chunk
+            del evals
             gc.collect(generation=2)
 
+        evalues = dt.rbind(*evalues)
+        evalues = evalues[:1, ["eVal"], dt.by(dt.f.subjectId), dt.sort(dt.f.eVal)]
+        # evalues.key = "subjectId"
+        # TODO: do a binary search to speed up the process
+        if evalue_perc is None:
+            logging.info("Trying to find best filtering threshold to fit data")
+            for perc in np.arange(0, 1, evalue_perc_step)[1:]:
+                # logging.info(
+                #     f"Removing alignments within {int(perc * 100)}% of the best e-value"
+                # )
+                # evalues[:, "maxlogEvalue"] = (
+                #     -1 * perc * (np.log(evalues[:, "eVal"]) / np.log(10))
+                # )
+                # evalues.key = "subjectId"
+                # aln_sizes = []
+                # for i in np.arange(len(alns)):
+                #     alns[i] = alns[i][:, :, dt.join(evalues)]
+                #     alns[i][:, "logEvalue"] = -1 * (
+                #         (np.log(alns[i][:, "eVal"])) / np.log(10)
+                #     )
+                #     aln_sizes.append(
+                #         alns[i][(dt.f.logEvalue > dt.f.maxlogEvalue), :].shape[0]
+                #     )
+                #     del alns[i]["maxlogEvalue"]
+                #     del alns[i]["logEvalue"]
+                aln_sizes = filter_eval_perc(alns=alns, evalues=evalues, perc=perc)
+                if np.sum(aln_sizes) < max_rows:
+                    for i in np.arange(len(alns)):
+                        alns[i] = alns[i][:, :, dt.join(evalues)]
+                        alns[i][:, "logEvalue"] = -1 * (
+                            (np.log(alns[i][:, "eVal"])) / np.log(10)
+                        )
+
+                        alns[i] = alns[i][(dt.f.logEvalue > dt.f.maxlogEvalue), :]
+
+                        del alns[i]["maxlogEvalue"]
+                        del alns[i]["logEvalue"]
+                        logging.info(f"Filter {perc} produced a manageable table.")
+                    break
+                else:
+                    logging.info(
+                        f"{np.sum(aln_sizes):,} alignments found. Increasing filtering threshold"
+                    )
+        else:
+            aln_sizes = filter_eval_perc(alns=alns, evalues=evalues, perc=evalue_perc)
+            if np.sum(aln_sizes) < max_rows:
+                for i in np.arange(len(alns)):
+                    alns[i] = alns[i][:, :, dt.join(evalues)]
+                    alns[i][:, "logEvalue"] = -1 * (
+                        (np.log(alns[i][:, "eVal"])) / np.log(10)
+                    )
+
+                    alns[i] = alns[i][(dt.f.logEvalue > dt.f.maxlogEvalue), :]
+
+                    del alns[i]["maxlogEvalue"]
+                    del alns[i]["logEvalue"]
+                logging.info(f"Filter {evalue_perc} produced a manageable table.")
+
+        if np.sum(aln_sizes) > max_rows:
+            logging.error(
+                f"The resulting table has more than {max_rows:,} alignments and it is not supported at the moment."
+            )
+            exit(1)
         logging.info(f"::: Concatenating chunks.")
         aln = alns[0]
         alns.pop(0)
@@ -122,13 +240,8 @@ def read_and_filter_alns(
         if len(alns) > 0:
             for i in alns:
                 aln.rbind(i)
-        if aln.shape[0] > max_rows:
-            logging.error(
-                f"The resulting table has more than {max_rows,} alignments and it is not supported at the moment."
-            )
-            exit(1)
     else:
-        logging.info(f"Pre-filtering: {nalns:,} alignments found")
+        logging.info(f"Pre-filtering: {nalns:,} alignments estimated")
         # logging.info(f"Read {nalns:,} alignments. Getting basic statistics.")
 
         logging.info(
@@ -290,48 +403,6 @@ def resolve_multimaps(
     return df1
 
 
-def get_stats_coverage(x, gen_data, trim=False, strim_5=18, strim_3=18):
-    """Get coverage statistics
-
-    Args:
-        x (list): A list with the coverage per position
-        strim_5 (int, optional): How much to trim on the 5'. Defaults to 18.
-        strim_3 (int, optional): How much to trim on the 3'. Defaults to 18.
-
-    Returns:
-        [type]: [description]
-    """
-    cov = gen_data[x]["cov"]
-    length = cov.shape[0]
-
-    # If we need to trim
-    if trim:
-        long_enough = cov.shape[0] >= strim_5 + strim_3 + 10
-        if long_enough:
-            cov = cov[strim_5:-strim_3]
-
-    cov_mean = cov.mean()
-    if cov_mean > 0:
-        cov_std = cov.std()
-        cov_evenness = cov_std / cov_mean
-        breadth = ((cov != 0).sum()) / length
-        breadth_exp = 1 - np.exp(-cov_mean)
-        if breadth >= breadth_exp:
-            breadth_exp_ratio = 1.0 * breadth
-        else:
-            breadth_exp_ratio = (breadth / breadth_exp) * breadth
-        return (
-            x,
-            cov_mean,
-            cov_std,
-            cov_evenness,
-            breadth,
-            breadth_exp,
-            breadth_exp_ratio,
-            gen_data[x]["n_alns"],
-        )
-
-
 def cov_stats(alns, refs):
 
     rle = alns.to_rle(nb_cpu=1)
@@ -358,7 +429,7 @@ def cov_stats(alns, refs):
     return df.drop("Run Value Sum Intermediate".split(), axis=1)
 
 
-def get_stats_coverage(x, gen_data, trim=False, strim_5=18, strim_3=18):
+def get_stats_coverage(x, gen_data, rl, trim=False, strim_5=18, strim_3=18):
     """Get coverage statistics
 
     Args:
@@ -371,13 +442,19 @@ def get_stats_coverage(x, gen_data, trim=False, strim_5=18, strim_3=18):
     """
     cov = gen_data[x]["cov"]
     length = cov.shape[0]
-
     # If we need to trim
     if trim:
         long_enough = cov.shape[0] >= strim_5 + strim_3 + 10
         if long_enough:
             cov = cov[strim_5:-strim_3]
 
+    if gen_data[x]["n_alns"] > 1:
+        rl_mean = np.mean(rl)
+        rl_std = np.std(rl, ddof=1)
+    else:
+        rl_mean = rl[0]
+        rl_std = 0
+
     cov_mean = cov.mean()
     if cov_mean > 0:
         cov_std = cov.std()
@@ -397,6 +474,8 @@ def get_stats_coverage(x, gen_data, trim=False, strim_5=18, strim_3=18):
             breadth_exp,
             breadth_exp_ratio,
             gen_data[x]["n_alns"],
+            rl_mean,
+            rl_std,
         )
 
 
@@ -404,6 +483,7 @@ def get_coverage_stats(df, trim=True):
 
     queries = defaultdict(int)
     gen_data = defaultdict(dict)
+    al = defaultdict(list)
     rl = defaultdict(list)
     # for a, q, b, c, d, e in tqdm.tqdm(
     #     zip(
@@ -432,30 +512,33 @@ def get_coverage_stats(df, trim=True):
         b = df[i, "subjectStart"]
         c = df[i, "subjectEnd"]
         d = df[i, "slen"]
-        e = df[i, "qlen"]
-
+        e = df[i, "alnLength"]
+        f = df[i, "qlen"]
         b = b - 1
 
         if a in gen_data:
             gen_data[a]["cov"][b:c] += 1
             gen_data[a]["n_alns"] += 1
-            rl[a] += [e / 3]
+            al[a] += [e / 3]
+            rl[a] += [f]
         else:
             gen_data[a]["cov"] = np.zeros(d, dtype=int)
             gen_data[a]["cov"][b:c] += 1
             gen_data[a]["n_alns"] = 0
             gen_data[a]["n_alns"] += 1
-            rl[a] += [e / 3]
+            al[a] += [e / 3]
+            rl[a] += [f]
     logging.info(
-        f"References will be dynamically trimmed at 5'/3'-ends (half of the avg. read length)"
+        f"References will be dynamically trimmed at 5'/3'-ends (half of the avg. aln length)"
     )
     stats = [
         get_stats_coverage(
             chrom,
             gen_data=gen_data,
-            strim_5=int(np.mean(rl[chrom]) / 2),
-            strim_3=int(np.mean(rl[chrom]) / 2),
+            strim_5=int(np.mean(al[chrom]) / 2),
+            strim_3=int(np.mean(al[chrom]) / 2),
             trim=trim,
+            rl=rl[chrom],
         )
         for chrom in tqdm.tqdm(
             gen_data,
@@ -479,6 +562,8 @@ def get_coverage_stats(df, trim=True):
         "breadth_expected",
         "breadth_expected_ratio",
         "n_alns",
+        "avg_read_length",
+        "stdev_read_length",
     ]
     return stats
 
@@ -504,12 +589,49 @@ def aggregate_gene_abundances(mapping_file, gene_abundances, threads=1):
     if mappings.shape[0] == 0:
         return None, None
     else:
-        mappings_agg = mappings.groupby("group").agg(
-            {"depth_mean": ["mean", "std", "median", "sum", "count"]}
+        mappings_agg_dm = mappings.groupby("group").agg(
+            {
+                "depth_mean": ["mean", "std", "median", "sum", "count"],
+            }
+        )
+        mappings_agg_dm = mappings_agg_dm.xs(
+            "depth_mean",
+            axis=1,
+            drop_level=True,
+        )
+        mappings_agg_dm = mappings_agg_dm.reset_index("group")
+        mappings_agg_dm.rename({"count": "n_genes"}, axis=1, inplace=True)
+
+        mappings_agg_al = mappings.groupby("group").agg(
+            {
+                "avg_read_length": ["mean"],
+            }
+        )
+        mappings_agg_al = mappings_agg_al.xs(
+            "avg_read_length",
+            axis=1,
+            drop_level=True,
+        )
+        mappings_agg_al = mappings_agg_al.reset_index("group")
+        mappings_agg_al.rename({"mean": "avg_read_length"}, axis=1, inplace=True)
+
+        mappings_agg_sl = mappings.groupby("group").agg(
+            {
+                "stdev_read_length": ["mean"],
+            }
+        )
+        mappings_agg_sl = mappings_agg_sl.xs(
+            "stdev_read_length",
+            axis=1,
+            drop_level=True,
+        )
+        mappings_agg_sl = mappings_agg_sl.reset_index("group")
+        mappings_agg_sl.rename({"mean": "stdev_read_length"}, axis=1, inplace=True)
+
+        mappings_agg = reduce(
+            lambda left, right: pd.merge(left, right, on=["group"], how="inner"),
+            [mappings_agg_dm, mappings_agg_al, mappings_agg_sl],
         )
-        mappings_agg = mappings_agg.xs("depth_mean", axis=1, drop_level=True)
-        mappings_agg = mappings_agg.reset_index("group")
-        mappings_agg.rename({"count": "n_genes"}, axis=1, inplace=True)
         return mappings, mappings_agg
 
 
